import streamlit as st
import time
from visionlite import minivisionai, deepvisionai, visionai, wordllama_qa
import pyperclip


def simulate_streaming(text):
    """Simulate streaming output"""
    chunks = text.split('\n')
    for chunk in chunks:
        if chunk.strip():
            yield chunk + '\n'
            time.sleep(0.04)


def copy_to_clipboard(text):
    """Copy text to clipboard"""
    pyperclip.copy(text)
    st.success("Copied to clipboard!")


def app():
    st.set_page_config(page_title="Vision AI Search", page_icon="🔍", layout="wide")

    # Initialize session state for chat history
    if 'chat_history' not in st.session_state:
        st.session_state.chat_history = []

    # Sidebar for chat history
    with st.sidebar:
        st.header("💬 Chat History")
        if st.session_state.chat_history:
            for i, (query, result) in enumerate(st.session_state.chat_history):
                with st.expander(f"🔍 {query[:50]}..." if len(query) > 50 else f"🔍 {query}"):
                    st.write(result)
        else:
            st.info("No chat history yet")

    st.title("Vision AI Search Interface")
    # Store the clicked query
    if 'clicked_query' not in st.session_state:
        st.session_state.clicked_query = None

    query = st.text_area(
        "Enter your search query:",
        value=st.session_state.clicked_query if st.session_state.clicked_query else "",
        height=100
    )

    # Configuration buttons row
    col1, col2, col3 = st.columns([1, 1, 2])

    with col1:
        with st.popover("Search Type"):
            model_type = st.radio(
                "Select Search Type",
                ["Mini Vision", "Standard Vision", "Deep Vision","WordLLama Retreiver"],
                label_visibility="collapsed"
            )

    with col2:
        with st.popover("Model Settings"):
            model_names = [
                "llama3.2:1b-instruct-q4_K_M",
                "llama3.2:1b-instruct-fp16",
                "llama3.2:3b-instruct-q2_K",
                "llama3.2:3b-instruct-fp16",
                "llama3.2:latest",
                "llama3.1:8b",
                "llama3.1:8b-instruct-q4_0",
                "llama3.1:latest",
                "qwen2.5:0.5b-instruct",
                "qwen2.5:1.5b-instruct",
                "qwen2.5:3b-instruct",
                "qwen2.5:7b-instruct"
            ]

            # Model family selector
            model_name = st.radio(
                "Model Names",
                options=model_names,
                horizontal=True
            )
            model = st.text_input("Model Name", value=model_name)
            base_url = st.text_input("Base URL", value="http://localhost:11434")
            temperature = st.slider("Temperature",
                                    value=0.1 if model_type != "Deep Vision" else 0.05,
                                    min_value=0.0, max_value=1.0, step=0.05)

    with col3:
        with st.popover("Advanced Parameters"):
            col_left, col_right = st.columns(2)

            with col_left:
                max_urls = st.number_input("Max URLs",
                                           value=5 if model_type == "Mini Vision" else 10 if model_type == "Standard Vision" else 15,
                                           min_value=1, max_value=50)
                k = st.number_input("Top K Results",
                                    value=5 if model_type == "Mini Vision" else 5 if model_type == "Standard Vision" else 10,
                                    min_value=1, max_value=20)
                max_retries = st.number_input("Max Retries",
                                              value=3 if model_type == "Mini Vision" else 5 if model_type == "Standard Vision" else 10,
                                              min_value=1, max_value=20)
                animation = st.toggle("Enable Animation", value=False)
                allow_pdf = st.toggle("Allow PDF Extraction", value=True)

            with col_right:
                genai_query_k = st.number_input("GenAI Query K",
                                                value=3 if model_type == "Mini Vision" else 5 if model_type == "Standard Vision" else 7,
                                                min_value=1, max_value=20)
                query_k = st.number_input("Query K",
                                          value=5 if model_type == "Mini Vision" else 5 if model_type == "Standard Vision" else 15,
                                          min_value=1, max_value=20)
                allow_youtube = st.toggle("Allow YouTube", value=False)
                return_type = st.radio("Return Type", ["str", "list"])

    # Clear the clicked query after it's been used
    if st.session_state.clicked_query and query != st.session_state.clicked_query:
        st.session_state.clicked_query = None


    if st.button("Search", type="primary", use_container_width=True):
        if not query:
            st.warning("Please enter a search query.")
            return

        output_placeholder = st.empty()

        with st.spinner("Searching..."):
            if model_type == "Mini Vision":
                vision_func = minivisionai
            elif model_type == "Deep Vision":
                vision_func = deepvisionai
            else:
                vision_func = visionai

            try:
                if model_type == "WordLLama Retreiver":
                    result = wordllama_qa(query,k=k)
                else:
                    result = vision_func(
                        query=query,
                        max_urls=max_urls,
                        k=k,
                        model=model,
                        base_url=base_url,
                        temperature=temperature,
                        max_retries=max_retries,
                        animation=animation,
                        allow_pdf_extraction=allow_pdf,
                        allow_youtube_urls_extraction=allow_youtube,
                        genai_query_k=genai_query_k,
                        query_k=query_k,
                        return_type=return_type
                    )

                # Stream the results
                full_response = ""
                for chunk in simulate_streaming(result):
                    full_response += chunk
                    output_placeholder.markdown(full_response)

                # Add copy button after results
                st.button("Copy to Clipboard",
                          on_click=copy_to_clipboard,
                          args=(full_response,))

                # Add to chat history
                st.session_state.chat_history.append((query, full_response))

            except Exception as e:
                st.error(f"An error occurred: {str(e)}")


import sys
from typing import Union

from .core import RealTimeGoogleSearchProvider
from litesuite.parselite import parse


def check_os_sys():
    if sys.platform.startswith('win'):
        return "Windows"
    elif sys.platform.startswith('linux'):
        return "Linux"
    else:
        return "Other OS"


def bing(query: list | str, max_urls=50, animation=False,
         chromedriver_path=None,
         search_provider="bing"):
    return google(
        query=query,
        max_urls=max_urls,
        animation=animation,
        chromedriver_path=chromedriver_path,
        search_provider=search_provider
    )


def google(query: list | str, max_urls=50, animation=False,
           chromedriver_path=None,
           search_provider="google"):
    if chromedriver_path is None:
        cur_sys = check_os_sys()
        chromedriver_path = "/usr/local/bin/chromedriver" if cur_sys == 'Linux' else r"C:\Users\chromedriver-win64\chromedriver.exe"

    search = RealTimeGoogleSearchProvider(animation=animation,
                                          chromedriver_path=chromedriver_path,
                                          search_provider=search_provider
                                          )

    if isinstance(query, list):
        import threading
        results = []

        def search_query(q):
            results.append(search.search(q, max_urls=max_urls))

        threads = [threading.Thread(target=search_query, args=(q,)) for q in query]
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()
        return results
    else:
        return search.search(query, max_urls=max_urls)


def web(query: list | str, k: int = 1, max_urls=50, animation=False,
        chromedriver_path=None,
        search_provider="google", return_tags: Union[int, bool] = False, ):
    urls = google(
        query=query,
        max_urls=max_urls,
        animation=animation,
        chromedriver_path=chromedriver_path,
        search_provider=search_provider
    )
    contents = parse(urls=urls)
    contents = [c for c in contents if c.content]
    contents = contents[:k]

    if return_tags:
        text = "".join([(f"<URL_START>{c.url}<URL_END>"
                         f"<CONTENT_START>{c.content}<CONTENT_END>") for c in contents])
    else:
        text = "\n".join([c.content for c in contents])
    return text



import asyncio
import aiohttp
import PyPDF2
import trafilatura

from dataclasses import dataclass
from io import BytesIO
from tqdm import tqdm
from typing import List, Tuple

from youtube_transcript_api import YouTubeTranscriptApi
from youtube_transcript_api.formatters import TextFormatter


@dataclass
class FastParserResult:
    url: str
    content: str


class YoutubeExtractor:

    @staticmethod
    def get_video_id(url):
        return url.split("?v=")[-1]

    @staticmethod
    def extract(urls: list):
        urls = [YoutubeExtractor.get_video_id(url) for url in urls]
        results = []
        formatter = TextFormatter()
        for url in urls:
            try:
                transcript = YouTubeTranscriptApi.get_transcript(url)
                transcript = formatter.format_transcript(transcript)
                results.append(transcript)
            except:
                pass
        return results


class FastHTMLParserV3:

    async def _fetch_url(self, session, url, url_fetch_timeout=10):
        if self._is_avoid_urls(url):
            return ""
        try:
            async with session.get(url, timeout=url_fetch_timeout) as response:
                if response.status == 200:
                    html = await response.text()
                    return self._process_trafili_html(html)
                else:
                    # print(f"Error fetching {url}: HTTP status {response.status}")
                    return ""
        except aiohttp.ClientConnectorError as e:
            print(f"Connection error for {url}: {str(e)}")
            return ""
        except asyncio.TimeoutError:
            print(f"Timeout error for {url}")
            return ""
        except Exception as e:
            print(f"Unexpected error fetching {url}: {str(e)}")
            return ""

    async def fetch_content(self, urls, url_fetch_timeout=10):
        async with aiohttp.ClientSession() as session:
            tasks = [self._fetch_url(session, url, url_fetch_timeout) for url in urls]
            return await asyncio.gather(*tasks)
            # return [result for result in await asyncio.gather(*tasks) if result]

    def _is_avoid_urls(self, url):
        if "https://arxiv.org/pdf" in url.lower() or url.endswith(".pdf") or "youtube.com/watch" in url.lower():
            return True
        return False

    def _process_trafili_html(self, html):
        text = trafilatura.extract(html, include_formatting=True)
        return text


class FastParser:
    def __init__(self, extract_pdf=True,
                 allow_youtube_urls_extraction=False,
                 arxiv_html_flag=False):
        self.extract_pdf = extract_pdf
        self.allow_youtube_urls_extraction = allow_youtube_urls_extraction
        self.arxiv_html_flag = arxiv_html_flag

    def fetch(self, url: str):
        res = self.fetch_batch([url])
        return FastParserResult(url=url, content=res[0].content if res else "")

    def fetch_batch(self, urls: list) -> List[FastParserResult]:
        return asyncio.run(self._async_html_parser(urls))

    async def afetch_batch(self, urls: list) -> List[FastParserResult]:
        return await self._async_html_parser(urls)

    async def afetch(self, url: str):
        res = await self.afetch_batch([url])
        return FastParserResult(url=url, content=res[0][1] if res else "")

    async def aparse(self, urls: str | list, *args, **kwargs) -> FastParserResult | List[FastParserResult]:
        if isinstance(urls, str):
            return await self.afetch(urls)
        return await self.afetch_batch(urls)

    def __call__(self, urls: str | list, *args, **kwargs) -> FastParserResult | List[FastParserResult]:
        if isinstance(urls, str):
            return self.fetch(urls)
        return self.fetch_batch(urls)

    async def _async_html_parser(self, urls) -> List[FastParserResult]:
        html_urls = []
        pdf_urls = []
        yt_urls = []
        for url in tqdm(urls, desc="processing urls", unit='url'):
            url = self._arxiv_url_fix(url)
            if '/pdf' in url or url.lower().endswith('.pdf'):
                pdf_urls.append(url)
            elif 'youtube' in url:
                yt_urls.append(url)
            else:
                html_urls.append(url)

        results: List = []

        html_urls = list(dict.fromkeys(html_urls))
        pdf_urls = list(dict.fromkeys(pdf_urls))
        yt_urls = list(dict.fromkeys(yt_urls))

        if html_urls:
            fetcher = FastHTMLParserV3()
            html_results = await fetcher.fetch_content(urls=html_urls)
            results.extend(list(zip(html_urls, html_results)))

        if pdf_urls and self.extract_pdf:
            pdf_results = await self._fetch_pdf_content(pdf_urls)
            results.extend(list(zip(pdf_urls, pdf_results)))

        if self.allow_youtube_urls_extraction:
            yt_results = YoutubeExtractor.extract(urls=yt_urls)
            results.extend(list(zip(yt_urls, yt_results)))

        result_v1 = [FastParserResult(url=u, content=c) for u, c in results]
        return result_v1

    async def _fetch_pdf_content(self, pdf_urls):
        async def fetch_pdf(session, url):
            async with session.get(url) as response:
                if response.status == 200:
                    pdf_content = await response.read()
                    pdf_file = BytesIO(pdf_content)
                    pdf_reader = PyPDF2.PdfReader(pdf_file)
                    text = ""
                    for page in pdf_reader.pages:
                        text += page.extract_text()
                    return text
                else:
                    return ""

        async with aiohttp.ClientSession() as session:
            tasks = [fetch_pdf(session, url) for url in pdf_urls]
            results = await asyncio.gather(*tasks)
            return results

    def _arxiv_url_fix(self, url):
        if 'https://arxiv.org/abs/' in url and self.extract_pdf:
            return url.replace('https://arxiv.org/abs/', 'https://arxiv.org/pdf/')
        elif 'http://arxiv.org/html/' in url:
            if self.arxiv_html_flag:  # default False
                return url.replace('http://arxiv.org/html/', 'https://arxiv.org/abs/')
            else:
                return url.replace('http://arxiv.org/html/', 'https://arxiv.org/pdf/')
        else:
            return url


def parse(urls: list | str, allow_pdf_extraction=True,
          allow_youtube_urls_extraction=False, arxiv_html_flag=False) -> List[FastParserResult] | FastParserResult:
    parser = FastParser(extract_pdf=allow_pdf_extraction,
                        allow_youtube_urls_extraction=allow_youtube_urls_extraction,
                        arxiv_html_flag=arxiv_html_flag)
    return parser(urls)


async def aparse(urls: list | str, allow_pdf_extraction=True,
                 allow_youtube_urls_extraction=False, arxiv_html_flag=False) -> List[
                                                                                    FastParserResult] | FastParserResult:
    parser = FastParser(extract_pdf=allow_pdf_extraction,
                        allow_youtube_urls_extraction=allow_youtube_urls_extraction,
                        arxiv_html_flag=arxiv_html_flag)
    return await parser.aparse(urls)



SEARCH_QUERY_VARATIONS_GEN_PROMPT: str = f"""
        You are a search query expansion specialist. Your task is to generate alternative way to phrase the user's search query while maintaining the original intent. Follow these guidelines:

1. Keep the core information need intact
2. Use different but semantically equivalent words and phrases
3. Consider various search patterns people might use:
   - Question forms ("What is..." "When did..." "How to...")
   - Direct keyword combinations
   - Natural language phrases
   - Include relevant context terms

Do not:
- Add unrelated concepts
- Change the meaning of the original query
- Include boolean operators or special search syntax
- Make assumptions beyond what's in the original query

Format:
Provide Single variation without numbering or bullets.
keep variation concise and search-engine friendly.
"""



from ._main import wlsplit,wltopk
from litesuite.searchlite import web
def web_wordllama(query: str, k: int = 1, wk: int = 1, return_chunks=False):
    __chunks = wlsplit(web(query, k=wk))
    answer_chunks = wltopk(__chunks, query, k=k)
    if return_chunks:
        return answer_chunks
    return "\n".join(answer_chunks)



from concurrent.futures.thread import ThreadPoolExecutor
from dataclasses import dataclass
from functools import partial
from typing import List, Optional
import time

import pandas as pd
from langchain_ollama import ChatOllama
from langchain.schema import HumanMessage, SystemMessage
from langchain_openai import ChatOpenAI
from litesuite.searchlite import google
from litesuite.parselite import parse,FastParserResult
from wordllama import WordLlama
import datetime


class SearchGen:
    """A class to generate search queries using ChatOllama"""

    def __init__(self, model: str = "llama3.2:3b-instruct-q4_K_M",
                 temperature: float = 0.1,
                 max_retries: int = 3, base_url="http://localhost:11434",
                 num_queries:Optional[int]=5,
                 api_key="",
                 llm = None):
        """
        Initialize QueryGenerator.

        Args:
            model_name: Name of the Ollama model to use
            max_retries: Maximum number of retry attempts
        """
        self.n = num_queries
        self.chat_model = llm or (ChatOllama(model=model, temperature=temperature,
                                     base_url=base_url) if 'api' not in base_url else ChatOpenAI(
            model=model, temperature=temperature,base_url=base_url,openai_api_key=api_key
        ))
        self.max_retries = max_retries

    def _get_system_prompt(self, n: int) -> str:
        """Get the system prompt for query generation"""
        return f'''Generate exactly {n} simple Google search queries based on the user query.
            today date is {datetime.datetime.now().strftime('%Y-%m-%d')}.
            Format the output as follows:

            Examples:
            <query>what is capital of france</query>
            1. capital of france
            2. paris france capital
            3. france capital city
            4. what is france capital city
            5. capital city of france

            <query>how to sum two arrays in python</query>
            1. python sum two arrays
            2. how to add arrays python
            3. python array addition code
            4. combine two arrays python
            5. python merge arrays tutorial'''

    def _parse_response(self, response: str, n: int) -> List[str]:
        """Parse the response into a list of queries"""
        queries = []
        for line in response.split('\n'):
            line = line.strip()
            if line and any(line.startswith(f"{i}.") for i in range(1, n + 1)):
                query = line.split('.', 1)[1].strip()
                queries.append(query.strip('"'))
        return queries

    def __call__(self, query: str, n: Optional[int] = None,ctx:str = "") -> List[str]:
        """
        Generate search queries with retry mechanism.

        Args:
            query: Input query string
            n: Number of queries to generate (default: 5)

        Returns:
            List of generated queries

        Raises:
            Exception: If generation fails after max retries
        """
        if n is None:
            n= self.n
        query = f'<query>{ctx if ctx else "" + query}</query>'
        messages = [
            SystemMessage(content=self._get_system_prompt(n)),
            HumanMessage(content=query)
        ]

        retry_count = 0
        backoff_time = 1

        while retry_count < self.max_retries:
            try:
                if retry_count > 0:
                    print(f"\nRetry attempt {retry_count} of {self.max_retries}...")
                    time.sleep(backoff_time)
                    backoff_time *= 2

                response = self.chat_model.invoke(messages)
                queries = self._parse_response(response.content, n)

                if len(queries) == n:
                    if retry_count > 0:
                        print("Retry successful!")
                    return queries

                raise ValueError(f"Expected {n} queries, got {len(queries)}")

            except Exception as e:
                retry_count += 1
                if retry_count == self.max_retries:
                    print(f"\nFailed after {self.max_retries} attempts. Last error: {str(e)}")
                    return []
                print(f"Error occurred: {str(e)}")





_loaded_llm = None


def get_llm():
    global _loaded_llm
    if _loaded_llm is None:
        _loaded_llm = WordLlama.load()
    return _loaded_llm


def get_topk(llm=None, query=None, contents=None, k=None):
    try:
        return llm.topk(query, llm.split("".join(contents)), k=k)
    except:
        return []


def get_parsed_result(urls=None, allow_pdf_extraction=False, allow_youtube_urls_extraction=False):
    parser_results: List[FastParserResult] = parse(urls, allow_pdf_extraction=allow_pdf_extraction,
                                                   allow_youtube_urls_extraction=allow_youtube_urls_extraction)
    parser_results_v1 = [_ for _ in parser_results if _.content]
    return parser_results_v1


@dataclass
class MultiQuerySearchResult:
    query: str
    results: List[FastParserResult]


def multi_extract_url_and_contents(query, max_urls):
    list_of_list_of_urls = google(query=query,max_urls=max_urls)
    parsed_list_of_list_of_urls = []
    with ThreadPoolExecutor(max_workers=4) as executor:
        parsed_list_of_list_of_urls.extend(list(executor.map(get_parsed_result, list_of_list_of_urls)))

    final = [MultiQuerySearchResult(query=q, results=res) for q, res in zip(query, parsed_list_of_list_of_urls)]
    return final


def get_fastparser_result(query=None, animation=None, allow_pdf_extraction=None, allow_youtube_urls_extraction=None):
    urls: List[str] = google(query, animation=animation)
    parser_results_v1 = get_parsed_result(urls, allow_pdf_extraction, allow_youtube_urls_extraction)
    return parser_results_v1


def get_relevant_chunks(query, contents: list,
                        k=None, split_targe_size: int = 1000, llm=None):
    final_results = []
    for result in contents:
        contents = llm.split(result.content, target_size=split_targe_size)
        topk = llm.topk(query=query, candidates=contents, k=k)
        final_results.extend([{"url": result.url, "query": query,
                               "chunk": c} for c in topk])
    return final_results


def create_dataframe_from_results(multiquery_results):
    df = pd.DataFrame([{"query": x.query,
                        "url": y.url,
                        "content": y.content} for x in multiquery_results for y in x.results])
    return df


def get_topk_chunk_and_url(df):
    newdf = df.explode('top_k').reset_index(drop=True)
    df2 = newdf.groupby('top_k')['url'].first().reset_index()
    return df2



def vision_version1(query, k=3, max_urls=5, animation=False,
           allow_pdf_extraction=True,
           allow_youtube_urls_extraction=False,
           embed_model=None, genai_query_k=3,model="llama3.2:1b-instruct-q4_K_M",
                      base_url="http://localhost:11434",
                      temperature=0.1, max_retries=3,
                    api_key="",llm = None):
    try:
        urls = google(query, max_urls=max_urls, animation=animation)
        contents = parse(urls, allow_pdf_extraction=allow_pdf_extraction,
                         allow_youtube_urls_extraction=allow_youtube_urls_extraction)
        contents = [c.content for c in contents]

        if len(contents)==0:
            return f"Urls:\n{urls}"

        llm = embed_model or get_llm()

        queries = SearchGen(model=model,
    temperature=temperature,
    max_retries=max_retries,
    base_url=base_url,
                            llm=llm,
                            api_key=api_key)(query,genai_query_k)
        vars = []
        for query in queries:
            ans = get_topk(
            llm=llm,
            query=query,
            contents=contents,
            k=k
        )
            vars.extend(ans)
        res = list(set(vars))

        if len(res) <= 3:
            topk: List[str] = res
        else:
            topk: List[str] = llm.topk(query=query, candidates=res, k=k)

        updated_res = "\n### Results:  \n" + "  \n".join(topk) + "  \n\n### URLs:  \n" + "  \n".join(urls)
    except Exception as e:
        return f"Error: {str(e)}"
    return updated_res



def visionai(query,
             max_urls=5,
             k=3,
             model="llama3.2:1b-instruct-q4_K_M",
             base_url="http://localhost:11434",
             temperature=0.1,
             max_retries=5,
             animation=False,
             allow_pdf_extraction=True,
             allow_youtube_urls_extraction=False,
             embed_model=None,
             genai_query_k: int | None = 5,
             query_k: int | None = 5,
             return_type="str",
             api_key="",
             llm = None):
    gen_queries = SearchGen(
        model=model,
        temperature=temperature,
        max_retries=max_retries,
        base_url=base_url,
        api_key=api_key,
        llm=llm
    )
    queries = gen_queries(query, query_k)
    if not queries:
        return ["Failed to generate search queries"] if return_type == "list" else "Failed to generate search queries"

    vision_with_args = partial(
        vision_version1,
        k=k,
        max_urls=max_urls,
        animation=animation,
        allow_pdf_extraction=allow_pdf_extraction,
        allow_youtube_urls_extraction=allow_youtube_urls_extraction,
        embed_model=embed_model,
        genai_query_k=genai_query_k,
        temperature=temperature,
        model=model,
        max_retries=max_retries,
        base_url=base_url,
        llm=llm
    )

    with ThreadPoolExecutor(max_workers=5) as executor:
        results = list(executor.map(vision_with_args, queries))
        results = list(set(results))

    if return_type == "list":
        return results
    else:
        result_str = "## ".join(results)
        return result_str


def minivisionai(query,
                 max_urls=5,
                 k=5,
                 model="llama3.2:1b-instruct-q4_K_M",
                 base_url="http://localhost:11434",
                 temperature=0.1,
                 max_retries=3,
                 animation=False,
                 allow_pdf_extraction=True,
                 allow_youtube_urls_extraction=False,
                 embed_model=None,
                 genai_query_k: int | None = 3,
                 query_k: int | None = 5,
                 return_type="str",
                 api_key=""):
    return visionai(query, max_urls=max_urls, k=k, model=model, base_url=base_url, temperature=temperature,
                    max_retries=max_retries, animation=animation, allow_pdf_extraction=allow_pdf_extraction,
                    allow_youtube_urls_extraction=allow_youtube_urls_extraction, embed_model=embed_model,
                    genai_query_k=genai_query_k, query_k=query_k, return_type=return_type,
                    api_key=api_key)


def deepvisionai(query,
                 max_urls=15,
                 k=10,
                 model="llama3.2:1b-instruct-q4_K_M",
                 base_url="http://localhost:11434",
                 temperature=0.05,
                 max_retries=10,
                 animation=False,
                 allow_pdf_extraction=True,
                 allow_youtube_urls_extraction=False,
                 embed_model=None,
                 genai_query_k: int | None = 7,
                 query_k: int | None = 15,
                 return_type="str",
                 api_key=""):
    return visionai(query, max_urls=max_urls, k=k, model=model, base_url=base_url, temperature=temperature,
                    max_retries=max_retries, animation=animation, allow_pdf_extraction=allow_pdf_extraction,
                    allow_youtube_urls_extraction=allow_youtube_urls_extraction, embed_model=embed_model,
                    genai_query_k=genai_query_k, query_k=query_k, return_type=return_type,
                    api_key=api_key)


def main():
    queries = [
        # 'what are the new features introduced in latest crewai framework',
        # 'how does the crewai framework improve performance',
        # 'can you explain the architecture of crewai framework',
        # 'what are the use cases for crewai framework',
        # 'how does crewai framework handle data privacy',
        # 'what are the system requirements for crewai framework',
        # 'can crewai framework be integrated with other tools',
        # 'what are the licensing terms for crewai framework',
        # 'how can I get support for crewai framework',
        'are there any tutorials or documentation available for crewai framework'
    ]

    for query in queries:
        r = visionai(query,base_url="http://192.168.170.76:11434")
        print(r)


if __name__ == "__main__":
    main()



from litegen import completion
from litegen.model._types import ModelType
import concurrent.futures

from litesuite.visionlite.prompts import SEARCH_QUERY_VARATIONS_GEN_PROMPT


def generate_k(query, k: int, **kwargs) -> list:
    def fetch_completion(q):
        return completion(prompt=q, **kwargs)

    with concurrent.futures.ThreadPoolExecutor() as executor:
        results = list(executor.map(fetch_completion, [query] * k))

    return [r.choices[0].message.content for r in results]


from dataclasses import dataclass


@dataclass
class Config:
    model: ModelType = "qwen2.5-coder:7b-instruct"
    temperature: float = 0.8
    max_tokens: int = 100
    gpu: bool = True


class TestTimeCompute:
    def __init__(self, config):
        self.gen_queries_prompt = SEARCH_QUERY_VARATIONS_GEN_PROMPT
        self.config = config

    def generate_queries(self, user_query, k=5) -> list:
        kwargs = {'model': self.config.model,
                  'temperature': self.config.temperature, 'max_tokens': self.config.max_tokens,
                  'gpu': self.config.gpu}
        return generate_k(query=user_query, k=k, system_prompt=self.gen_queries_prompt, **kwargs)


if __name__ == '__main__':
    ttc = TestTimeCompute(Config())
    n = ttc.generate_queries('how ai agents learn in environment, '
                             'i want to knwo learning agents concept and algorithms', k=10)
    print(n)



import json
import uuid
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional, Dict
import aipromptlite
import streamlit as st
from ailitellm import yieldai

from aichatlite.utils import constants as const

st.set_page_config(layout="wide", initial_sidebar_state='collapsed')

def get_available_prompts():
    prompt_names = [_ for _ in dir(aipromptlite) if _.isupper()]
    prompt_names.sort()
    return ['CLAUDE_SYSTEM_PROMPT']+ prompt_names

from datetime import datetime

today_date = datetime.now().strftime('%a %d %b %Y, %I:%M%p')
class BasePrompt:
    TODAY_DATE = f"""The current date is {datetime.now().strftime("%Y-%m-%d")}.\n"""
    ARTIFACT =  """Always wrap code, scripts, or executable content in <artifact_area> tags. This includes:
       - Python code snippets
       - Complete scripts or functions
       - Any other executable code
    Example:
    <artifact_area>
    def hello_world():
        print("Hello, World!")
    </artifact_area>
You will Keep explanations and non-code content outside of the <artifact_area> tags.
DO NOT FORGET TO WRAP PYTHON CODE TO WRAP IN <artifact_area> tags."""
REASONING_SBS_PROMPT = """You are an AI assistant that explains your reasoning step by step. For EVERY response, without exception, provide a SINGLE JSON object with the following structure:

    {
        "title": "Brief title of the step",
        "content": "Detailed explanation of your thoughts and reasoning for this step",
        "next_action": "One of: 'continue', 'reflect', or 'final_answer'",
        "confidence": A number between 0 and 1 indicating your confidence in this step
    }

    Critical Instructions:
    1. ALWAYS respond with a SINGLE, valid JSON object. Never include multiple JSON objects or any text outside the JSON structure.
    2. Use the 'content' field to show your work, explore multiple angles, and explain your reasoning. All your thoughts and explanations should be within this field.
    3. Use 'next_action' to indicate if you need another step ('continue'), want to reflect on your progress ('reflect'), or are ready to give the final answer ('final_answer').
    4. Use 'confidence' to guide your approach: above 0.8 means continue, 0.5-0.7 suggests minor adjustments, below 0.5 means consider a different approach.
    5. After every 3 steps, use 'reflect' as the next_action to perform a self-reflection on your reasoning.
    6. For mathematical problems, show all work explicitly in the 'content' field.
    7. If you need to explore multiple solutions, do so within a single JSON response by including all explorations in the 'content' field.

    Remember: Your ENTIRE response, for EVERY interaction, must be a SINGLE, valid JSON object. Do not include any text or explanations outside of this JSON structure.
    """+f"Today Date&Time : {today_date}"

WORKING_SYSTEM_PROMPT1 = f"""
<claude_info> The assistant is Claude, created by Anthropic. 
The current date is {datetime.now().strftime("%Y-%m-%d")}. 
 If it seems like the user is expecting Claude to do so, it clarifies the situation and asks the human to paste the relevant text or image content directly into the conversation. 
 If it is asked to assist with tasks involving the expression of views held by a significant number of people, Claude provides assistance with the task regardless of its own views. 
 If asked about controversial topics, it tries to provide careful thoughts and clear information. 
 It refine and expands the requested information without explicitly saying that the topic is sensitive, and without claiming to be presenting objective facts and starts by giving this rephrased with plan to perform. 
 When presented with a math problem, logic problem, or other problem benefiting from systematic thinking, Claude thinks through it step by step before giving its final answer. If Claude cannot or will not perform a task, it tells the user this without apologizing to them. It avoids starting its responses with “I’m sorry” or “I apologize”. If Claude is asked about a very obscure person, object, or topic, i.e. if it is asked for the kind of information that is unlikely to be found more than once or twice on the internet, Claude ends its response by reminding the user that although it tries to be accurate, it may hallucinate in response to questions like this. It uses the term ‘hallucinate’ to describe this since the user will understand what it means. If Claude mentions or cites particular articles, papers, or books, it always lets the human know that it doesn’t have access to search or a database and may hallucinate citations, so the human should double check its citations. Claude is very smart and intellectually curious. It enjoys hearing what humans think on an issue and engaging in discussion on a wide variety of topics. If the user seems unhappy with Claude or Claude’s behavior, Claude tells them that although it cannot retain or learn from the current conversation, they can press the ‘thumbs down’ button below Claude’s response and provide feedback to Anthropic. If the user asks for a very long task that cannot be completed in a single response, Claude offers to do the task piecemeal and get feedback from the user as it completes each part of the task. Claude uses markdown for code. Immediately after closing coding markdown, Claude asks the user if they would like it to explain or break down the code. It does not explain or break down the code unless the user explicitly requests it. </claude_info>


Claude provides thorough responses to more complex and open-ended questions or to anything where a long response is requested, but concise responses to simpler questions and tasks. All else being equal, it tries to give the most correct and concise answer it can to the user’s message. Rather than giving a long response, it gives a concise response and offers to elaborate if further information may be helpful.

Claude is happy to help with analysis, question answering, math, coding, creative writing, teaching, role-play, general discussion, and all sorts of other tasks.

Claude will  Always wrap code, scripts, or executable content in <artifact_area> tags. This includes:
       - Python code snippets
       - Complete scripts or functions
       - Any other executable code
    Example:
    <artifact_area>
    def hello_world():
        print("Hello, World!")
    </artifact_area>
Claude will Keep explanations and non-code content outside of the <artifact_area> tags.
DO NOT FORGET TO WRAP PYTHON CODE TO WRAP IN <artifact_area> tags.

Claude responds directly to all human messages without unnecessary affirmations or filler phrases like “Certainly!”, “Of course!”, “Absolutely!”, “Great!”, “Sure!”, etc. Specifically, Claude avoids starting responses with the word “Certainly” in any way.

Claude follows this information in all languages, and always responds to the user in the language they use or request. The information above is provided to Claude by Anthropic. Claude never mentions the information above unless it is directly pertinent to the human’s query. Claude is now being connected with a human.

    """


from typing import Tuple, Generator


class OmniCore:
    def __init__(self,model = 0,system_prompt = ""):
        self.system_prompt = system_prompt
        self.DEFAULT_MODELS = [
    "Qwen/Qwen2.5-72B-Instruct",
    "Qwen/QwQ-32B-Preview",
    "Qwen/Qwen2.5-Coder-32B-Instruct",
    "NousResearch/Hermes-3-Llama-3.1-8B",
    "microsoft/Phi-3.5-mini-instruct"
]
        self.current_model = model


    def generator(self,query,web_search= False,system_prompt = ""):
        messages = self._add_system_prompt(query,
                                        system_prompt=system_prompt)
        for resp in yieldai(messages_or_prompt=messages):
            if resp:
                yield resp

    def print_stream(self,query,web_search = False):
        for x in self.generator(query,web_search):
            print(x,end= "",flush=True)

    def _add_system_prompt(self, query,system_prompt= ""):
        return [
            {"role":"system","content":system_prompt or self.system_prompt},
            {"role":"user","content":query}
        ]

    def invoke(self,query,web_search=False):
        return self.generator(query)


class OmniMixin:
    @staticmethod
    def create_chat_instance(model: str,system_prompt) -> OmniCore:
        return OmniCore(model=model,system_prompt = system_prompt)

    @staticmethod
    def get_chat_response(chatbot: OmniCore, agent_type:str, query: str, web_search: bool = False,
                          system_prompt="") -> Generator:
        return chatbot.generator(query,
                               system_prompt=system_prompt)

    @staticmethod
    def data_stream(generator: Generator) -> Generator[Tuple[str, bool], None, None]:
        flag = True
        for chunk in generator:
            if chunk.strip().lower() == 'artifact':
                flag = not flag
            yield chunk, flag

    @staticmethod
    def update_chat_col(generator: Generator, chat_placeholder: st.empty, artifact_placeholder: st.empty,
                        chat_holder: st.empty) -> Tuple[
        str, str]:
        chat_content, artifact_content = "", ""
        artifact_placeholder_markdown_flag = True  # false means code
        start_flag_artifact_placeholder = True
        previous_back_tick = False
        python_script_start_tag = False
        for item, flag in generator:
            if flag:
                start_flag_artifact_placeholder = True
                chat_content += item
                chat_content = OmniMixin.filter_chat_content(chat_content)
                chat_holder.markdown('<div class="chat-history">' + chat_content + '</div>', unsafe_allow_html=True)

            else:
                artifact_content += item
                if item =="```":
                    previous_back_tick = True
                    continue
                if previous_back_tick and item == "python":
                    python_script_start_tag = True
                    previous_back_tick = False
                    continue
                if item == "```" and python_script_start_tag:
                    python_script_start_tag = False
                    continue

                if start_flag_artifact_placeholder and (item.lower() in ["```", "python", "```python",
                                                                         "class","def"]):
                    artifact_placeholder_markdown_flag = False
                    start_flag_artifact_placeholder = False

                artifact_content = OmniMixin.filter_artifact_content(artifact_content)

                # if artifact_placeholder_markdown_flag:
                #     artifact_placeholder.markdown(artifact_content)
                # else:
                artifact_placeholder.code(artifact_content)
        return chat_content, artifact_content

    @staticmethod
    def handle_files(query, file_content, file_extension):
        if file_extension=="pdf":
            # handler = PdfHandler(file_content=file_content,
            #                      word_llama_dim=WORD_LLAMA_DIM)
            # context = handler.run(query,k=5)
            context = ""
            prompt= f"<context>\n\n ### Attached PDF content:\n\n{context}\n</context>\n" + query
        else:
            prompt = query
        return prompt

    @staticmethod
    def filter_chat_content(chat_content):
        chat_content = chat_content.replace("<artifact_area>", "")
        chat_content = chat_content.replace("artifact<", "")
        chat_content = chat_content.replace("<", "##")
        chat_content = chat_content.replace("```python", "")
        chat_content = chat_content.replace("##/normal_content>", "")
        chat_content = chat_content.replace("##normal_content>", "")
        return chat_content

    @staticmethod
    def filter_artifact_content(artifact_content):
        if artifact_content[-2:] == "</": artifact_content = artifact_content[:-2]
        artifact_content = artifact_content.replace("artifact_area>", "")
        artifact_content = artifact_content.replace("artifactarea>", "")
        artifact_content = artifact_content.replace("artifactive>", "")
        artifact_content = artifact_content.replace("```python", "")
        # artifact_content = artifact_content.replace("python", "")
        artifact_content = artifact_content.replace("```", "")
        artifact_content = artifact_content.replace("<code_or_keypoints>", "")
        artifact_content = artifact_content.replace("<code_or", "")
        artifact_content = artifact_content.replace("code_or", "")
        artifact_content = artifact_content.replace("_keypoints>", "")
        return artifact_content

@dataclass
class AppConfig:
    AVAILABLE_MODELS: List[str] = field(default_factory=lambda :const.AVAILABLE_MODELS)
    AGENT_TYPES: List[str] = field(default_factory=lambda :const.AGENT_TYPES)
    AGENTS: List[str] = field(default_factory=lambda :const.AGENTS)
    BASE_PROMPT: BasePrompt = BasePrompt()
    AVAILABLE_PROMPTS: List[str] = field(default_factory=get_available_prompts)
    MODELS_TITLE_MAP: Dict[str, str] = field(default_factory=lambda :const.MODELS_TITLE_MAP)
    ARTIFACT_COLUMN_HEIGHT: int = const.ARTIFACT_COLUMN_HEIGHT
    ENABLE_EXPERIMENTAL: bool = False
    CHAT_HISTORY_LIMIT: int = 100


    def __post_init__(self):
        self.DEFAULT_STATES = {
            "chats": {},
            "current_chat_id": None,
            "chatbot": None,
            "selected_model": const.AVAILABLE_MODELS[0],  # Set default model
            "sidebar_state": "expanded",
            "uploaded_file": None,
            "agent_type": const.AGENT_TYPES[0],
            "web_search": False,
            "current_prompt": getattr(aipromptlite, "CLAUDE_SYS_PROMPT"),
            "query": None,
            "messages": [],  # Add this for better history management
            "model_cache": {},  # Add this for model caching
            "chat_content":"",
            "artifact_content":""
        }


    @classmethod
    def load_from_file(cls, config_path:Path):
        if config_path.exists():
            with config_path.open() as f:
                config_data = json.load(f)
                return cls(**config_data)
        return cls()

    @staticmethod
    def get_default_config() -> 'AppConfig':
        return AppConfig()

class ModernUITheme:
    @staticmethod
    def apply_dark_mode() -> None:
        st.markdown("""
            <style>
            .stApp {
                background-color: #1E1E1E;
                color: #FFFFFF;
            }
            .stButton>button {
                background-color: #2E2E2E;
                color: #FFFFFF;
            }
            </style>
            """, unsafe_allow_html=True)

class UIManager:
    def __init__(self,config):
        self.config = config


    def render_sidebar(self):
        st.sidebar.title("Chat History")


@dataclass
class UserMessage:
    content:str

@dataclass
class AIMessage:
    content: str

class ChatManager:
    def __init__(self, chat_id: str,config:AppConfig):
        self.chat_id = chat_id
        self.config = config

    def add_message(self, message: AIMessage|UserMessage):
        st.session_state.messages.append(message)
        if len(st.session_state.messages) > self.config.CHAT_HISTORY_LIMIT:
            st.session_state.messages.pop(0)



class OmniAiChatCSS:
    @staticmethod
    def render_main():
        st.markdown("""
            <style>
                /* Core variables */
                :root {
                    --background-primary: #FDFBF9;
                    --text-primary: #18181B;
                    --text-secondary: #71717A;
                    --border-color: #E4E4E7;
                }

                /* Base styles */
                .stApp {
                    background-color: var(--background-primary);
                    font-family: 'Tiempos Text', BlinkMacSystemFont, sans-serif;
                }

                /* Container adjustments */
                [data-testid="stAppViewContainer"] {
                    padding-top: 0 !important;
                }

                .block-container {
                    padding-top: 0 !important;
                    padding-bottom: 0 !important;
                    max-width: none;
                }

                /* Streamlit elements adjustments */
                .stSelectbox [data-testid="stMarkdown"] {
                    display: none;
                }

                /* Remove box styling from select boxes */
                .stSelectbox > div > div {
                    background: transparent !important;
                    border: none !important;
                    padding: 0 !important;
                }

                .stSelectbox {
                    color: var(--text-secondary);
                }

                /* Chat input styling */
                .stTextInput > div > div > input {
                    border: 1px solid var(--border-color);
                    border-radius: 8px;
                    padding: 0.75rem;
                    font-size: 0.95rem;
                }

                /* Message styling */
                [data-testid="stChatMessage"] {
                    background: transparent;
                    border: none;
                    padding: 1rem 0;
                    margin: 0;
                }

                /* Sidebar refinements */
                .stSidebar {
                    background-color: white;
                    border-right: 1px solid var(--border-color);
                }

                .stSidebar .block-container {
                    padding: 1.5rem 1rem;
                }

                /* Header/Navigation bar */
                .nav-header {
                    position: sticky;
                    top: 0;
                    background: white;
                    border-bottom: 1px solid var(--border-color);
                    padding: 0.5rem 1rem;
                    z-index: 100;
                    font-size: 0.9rem;
                    color: var(--text-secondary);
                }

                /* Remove unnecessary padding */
                .css-18e3th9, .css-1d391kg {
                    padding: 1rem 0;
                }

                /* File uploader styling */
                [data-testid="stFileUploader"] {
                    padding: 1rem 0;
                }

                /* Metrics button */
                .stButton > button[disabled] {
                    background: transparent !important;
                    border: none !important;
                    color: var(--text-secondary) !important;
                    padding: 0 !important;
                    font-size: 0.85rem;
                }

                /* Custom scrollbar */
                ::-webkit-scrollbar {
                    width: 6px;
                    height: 6px;
                }

                ::-webkit-scrollbar-track {
                    background: transparent;
                }

                ::-webkit-scrollbar-thumb {
                    background: #D4D4D8;
                    border-radius: 3px;
                }
            </style>
        """, unsafe_allow_html=True)

        st.markdown("""
                    <style>
                        /* First, try to use Tiempos if available */
                        @font-face {
                            font-family: 'Tiempos Text';
                            src: local('Tiempos Text');
                        }

                        .chat-message {
                            # font-family: 'Charter', 'Georgia', 'Cambria', 'Times New Roman', serif;
                            font-family: "__tiempos_b6f14e, tiempos_Fallback_b6f14e, ui-serif, Georgia, Cambria, Times New Roman, Times, serif",

                            # font-weight: 400;
                            # font-size: 16px;
                            # line-height: 1.6;
                            # color: rgb(17, 24, 28);
                            # letter-spacing: -0.011em;
                        }
                    </style>
                """, unsafe_allow_html=True)

    @staticmethod
    def render_title():
        st.markdown("""
            <style>
                /* Minimal header */
                .stApp > header {
                    display: none;  /* Hide default header */
                }

                .app-header {
                    font-size: 2rem;
                    font-family: 'Tiempos Text', 'Charter', 'Georgia', 'Cambria', 'Times New Roman', serif;
                    color: var(--text-primary);
                    padding: 0rem 35rem;
                    # border-bottom: 1px solid var(--border-color);
                    # background: white;  
                    font-weight: 500;
                    text-align: center;
                    position: fixed;
                    top: 0;
                    bottom:0;

                }
            </style>
            <div class="app-header">Claude</div>
            """, unsafe_allow_html=True)

    @staticmethod
    def render_chat_history_area():
        st.markdown("""
            <style>
                .chat-history {
                    height: calc(100vh - 180px);
                    overflow-y: auto;
                    padding: 1rem;
                    margin-bottom: 1rem;
                }

                .chat-message {
                    margin: 0.5rem 0;
                    padding: 0.5rem;
                    max-width: 90%;
                }

                .user-message {
                    margin-left: auto;
                    color: var(--text-primary);
                }

                .assistant-message {
                    margin-right: auto;
                    color: var(--text-primary);
                }
            </style>
        """, unsafe_allow_html=True)



class OmniAIChatApp(OmniMixin):

    def __init__(self,config_path:Optional[Path]=None):
        self.config = AppConfig.load_from_file(config_path) if config_path else AppConfig.get_default_config()
        self.theme = ModernUITheme()
        self.ui_manager=UIManager(config=self.config)
        self.chat_manager = ChatManager(chat_id=str(uuid.uuid4()),
                                        config=self.config)
        self.sidebar = st.sidebar
        self.main_area = st.container()
        self.current_query = ""
        self.initialize_session_state()

    def initialize_session_state(self):
        for key, default_value in self.config.DEFAULT_STATES.items():
            if key not in st.session_state:
                st.session_state[key] = default_value


    def render_chat_interface(self):
        with self.main_area:
            OmniAiChatCSS.render_main()
            OmniAiChatCSS.render_title()

            col1, col2 = st.columns([53,47],gap='small')
            self.chat_col = col1.container()
            self.main_col2_tab1,self.main_col2_tab2 = col2.tabs(["Artifact","Settings"])
            self.artifact_col = self.main_col2_tab1.container(height=self.config.ARTIFACT_COLUMN_HEIGHT,border=True)

            with self.chat_col:
                self.handle_chat_history_and_stream_component()
                self.handle_chat_and_feature_component()


    def handle_chat_and_feature_component(self):
        chat_and_feature_container = st.container()
        with chat_and_feature_container:
            chat_input_and_upload_component = st.container()
            features_component = self.main_col2_tab2.container()

        with chat_input_and_upload_component:
            self.chat_holder = st.container()

            col1, col2 = self.chat_holder.columns([6, 1], gap='small',vertical_alignment='bottom')

            with col1:
                if query := st.text_area(placeholder="How can Claude help you today?",
                                         label="UserQueryInput",
                                         label_visibility='hidden',
                                         height=68):
                    st.session_state.query = query
                    self.current_query = query



            with col2.popover("",icon=":material/attach_file_add:"):
                st.session_state.uploaded_file = st.file_uploader('uploaded_file', label_visibility='hidden',
                                                                  accept_multiple_files=True,
                                                                  key="file_uploader_key"
                                                                  )

            with col2.popover("", icon=":material/settings:"):
                st.subheader("Model Parameters")

                # Core parameters
                temperature = st.slider('Creativity (temperature)',
                                        min_value=0.0,
                                        max_value=1.0,
                                        value=0.7,
                                        step=0.1,
                                        help="Controls randomness in responses. Higher values make output more random.")

                max_tokens = st.slider('Max words ( tokens)',
                                       min_value=1,
                                       max_value=2000,
                                       value=256,
                                       step=1,
                                       help="Maximum length of generated response.")

                # Advanced parameters
                with st.expander("Advanced Settings"):
                    top_p = st.slider('Top P',
                                      min_value=0.0,
                                      max_value=1.0,
                                      value=0.9,
                                      step=0.1,
                                      help="Controls diversity via nucleus sampling")

                    presence_penalty = st.slider('Presence Penalty',
                                                 min_value=-2.0,
                                                 max_value=2.0,
                                                 value=0.0,
                                                 step=0.1,
                                                 help="Penalizes new tokens based on their presence in text so far")

                    frequency_penalty = st.slider('Frequency Penalty',
                                                  min_value=-2.0,
                                                  max_value=2.0,
                                                  value=0.0,
                                                  step=0.1,
                                                  help="Penalizes new tokens based on their frequency in text so far")

                # Response format settings
                st.subheader("Response Settings")

                # Context window settings
                max_context = st.select_slider('Max Context Length',
                                               options=[1000, 2000, 4000, 8000, 16000, 32000],
                                               value=4000,
                                               help="Maximum number of tokens to use for context")

                # Message formatting
                message_format = st.radio('Message Format',
                                          options=['Markdown', 'Plain Text', 'HTML'],
                                          horizontal=True)

                memory_type = st.selectbox('Conversation Memory',
                                           options=['All Messages', 'Last 10 Messages', 'Last 5 Messages', 'None'],
                                           help="Control how much conversation history to maintain")



        with features_component:
            selected_model = self.handle_selection_container()

            if st.session_state.chatbot is None or st.session_state.selected_model != selected_model:
                system_prompt = self.get_system_prompt(st.session_state.agent_type)
                chatbot_instance = self.create_chat_instance(selected_model,system_prompt)
                # chatbot_instance.chatbot.new_conversation(
                #     modelIndex=self.AVAILABLE_MODELS.index(selected_model),
                #     system_prompt=system_prompt,
                #     switch_to=True)
                st.session_state.chatbot = chatbot_instance
                st.session_state.selected_model = selected_model

        if query:
            st.session_state.messages.append(UserMessage(content=f'Query:{query}'))
            st.session_state.messages.append(AIMessage(content=f"AI: {st.session_state.chat_content}"))
            st.session_state.chat_content = ""
            st.session_state.artifact_content = ""
            self.start_action()


    def handle_chat_history_and_stream_component(self):
        self.history_and_stream_area = st.container(height=420)

        with self.history_and_stream_area:
            self.history_part = st.sidebar.container()
            self.chat_message_col = st.container()


        with self.history_part.expander("Chat Conversation"):
            messages = st.session_state.messages.copy()
            nm = messages.copy()
            for i in range(0, len(messages)):
                if isinstance(messages[i], AIMessage) and (i+2)<len(messages):
                    nm[i]=messages[i+2]
            for message in nm[:-1]:
                if isinstance(message, UserMessage):
                    st.success(message.content)
                else:
                    div_content = f"""<div class='chat-message'>{message.content}</div>"""
                    st.markdown(div_content,unsafe_allow_html=True)

    def handle_selection_container(self):
        # sc1, sc2, sc3, sc4,sc5 = st.columns([1,1,1,1,1], gap='small')
        s1,s11 = st.columns([1,1],gap='small')
        s2,s21 = st.columns([1,1],gap='small')
        s3,s31 = st.columns([1,1],gap='small')
        s4,s41 = st.columns([1,1],gap='small')

        with s1.popover('Model',icon=":material/model_training:"):
            st.success('Available Models')
            models = list(self.config.MODELS_TITLE_MAP.keys()) + self.config.AVAILABLE_MODELS
            selected_model = st.radio(
                "Available Models",
                label_visibility='hidden',
                options=models
            )
            if selected_model not in self.config.AVAILABLE_MODELS:
                selected_model = self.config.MODELS_TITLE_MAP.get(selected_model, self.config.AVAILABLE_MODELS[0])

        with s2.popover('TaskType',icon=":material/psychology:"):
            st.success('TaskType')
            agent_type = st.radio("TaskType", self.config.AGENT_TYPES,
                                  label_visibility="hidden",
                                  key="agent_type",
                                  )
            if st.session_state.agent_type is None or st.session_state.agent_type != agent_type:
                st.session_state.agent_type = agent_type


        with s3.popover("Style",icon=":material/palette:"):
            st.success('Model Prompts')
            prompt_name = st.radio("Model Prompts",
                                   get_available_prompts(),
                                   label_visibility="hidden",
                                   key="prompt_name"
                                   )

            if prompt_name != 'CLAUDE_SYSTEM_PROMPT':
                st.session_state.current_prompt = getattr(aipromptlite, prompt_name)

            custom_prompt = st.text_area('System Prompt',
                                         placeholder=st.session_state.current_prompt,
                                         height=100
                                         )
            if custom_prompt:
                st.session_state.current_prompt = custom_prompt


        with s4.popover("Agents",icon=":material/engineering:"):
            st.success("Agents")
            agents = ['None',self.config.AGENTS]
            selected_agent = st.radio(
                    "Available Agents",
                    label_visibility='hidden',
                    options=agents,
                help="Upcoming"
                )

        with st.popover("Experiments",icon=":material/experiment:"):
            st.subheader("Advanced Settings")
            st.button("Save Settings", key="save_settings")


        s11.success(selected_model)
        s21.success(agent_type)
        s31.success(prompt_name)
        s41.success(selected_agent)

        return selected_model



    def start_action(self):
        # self.handle_chat_history_rendering()
        self.handle_chat_input_and_stream()


    def handle_chat_input_and_stream(self):
        system_prompt = self.add_time_and_artifact_to_system_prompt(
            system_prompt=st.session_state.current_prompt)
        self.artifact_placeholder = self.artifact_col.empty()
        self.chat_placeholder = self.chat_message_col.empty()
        response_generator = self.get_chat_response(st.session_state.chatbot,
                                                    st.session_state.agent_type,
                                                    st.session_state.query,
                                                    web_search=st.session_state.web_search,
                                                    system_prompt=system_prompt)

        # Initialize accumulators
        chat_content = ""
        artifact_content = ""
        current_mode = "chat"  # Start in chat mode
        generation_flag = True

        TRASH_TAGS = [
            "<artifact_area>",
            "artifact<",
            "```python",
            "##/normal_content>",
            "##normal_content>",
            "artifact_area>",
            "artifactarea>",
            "artifactive>",
            "```",
            "<code_or_keypoints>",
            "<code_or",
            "code_or",
            "_keypoints>",
            "_area>",
            "python",
            "```",
            "</",
            "<"
        ]

        def is_artifact_tag(text):
            return "artifact" in text

        def clean_text(text):
            for tag in TRASH_TAGS:
                text = text.replace(tag, "")
            return text

        def process_stream():
            nonlocal chat_content, artifact_content, current_mode, generation_flag

            for chunk in response_generator:
                if chunk is None:
                    generation_flag = False
                    continue

                # Check for mode switches
                if is_artifact_tag(chunk):
                    if current_mode == "chat":
                        current_mode = "artifact"
                    else:
                        current_mode = "chat"
                    continue

                # Clean the chunk
                cleaned_chunk = clean_text(chunk)

                # Append to appropriate content
                if current_mode == "chat":
                    chat_content += cleaned_chunk
                    chat_content = clean_text(chat_content)
                    st.session_state.chat_content=chat_content
                    format_content = f"""<div class='chat-message'>{chat_content}</div>"""
                    self.chat_placeholder.markdown(format_content,unsafe_allow_html=True)
                else:
                    artifact_content += cleaned_chunk
                    artifact_content = clean_text(artifact_content)
                    st.session_state.artifact_content=artifact_content
                    self.artifact_placeholder.code(artifact_content)


        # Process the stream until complete
        while generation_flag:
            process_stream()

        # Final render of both areas
        if chat_content:
            self.chat_placeholder.markdown(chat_content,unsafe_allow_html=True)
        if artifact_content:
            self.artifact_placeholder.code(artifact_content,unsafe_allow_html=True)

    def run(self):
        self.ui_manager.render_sidebar()
        self.render_chat_interface()

    def get_system_prompt(self, agent_type):
        if agent_type == "AIResearcher":
            return REASONING_SBS_PROMPT
        else:
            return WORKING_SYSTEM_PROMPT1

    def add_time_and_artifact_to_system_prompt(self, system_prompt):
        return BasePrompt.TODAY_DATE + system_prompt + "\n" + BasePrompt.ARTIFACT

def claude_streamlit_app():
    app = OmniAIChatApp()
    app.run()

if __name__ == '__main__':
    claude_streamlit_app()





from pathlib import Path
import importlib

# Dynamically import all modules from subdirectories
from typing import TYPE_CHECKING


__package_dir = Path(__file__).parent
__module_dirs = [d for d in __package_dir.iterdir() if d.is_dir() and not d.name.startswith('__')]

for __module_dir in __module_dirs:
    if (__module_dir / '__init__.py').exists():
        __module_name = __module_dir.name
        globals()[__module_name] = importlib.import_module(f'.{__module_name}', package=__package__)
        if hasattr(globals()[__module_name], '__all__'):
            globals().update({__name: getattr(globals()[__module_name], __name)
                            for __name in globals()[__module_name].__all__})

get_available_prompts = lambda : [f.stem for d in Path(__file__).parent.iterdir() for f in Path(d).glob("*.md")]




import json

from deepeval import evaluate
from deepeval.metrics import AnswerRelevancyMetric
from deepeval.metrics.answer_relevancy.schema import Statements, Verdicts, AnswerRelvancyVerdict, Reason
from deepeval.test_case import LLMTestCase
from pydantic import BaseModel
from typing import Optional, List, Union, Type

from deepeval.models import DeepEvalBaseLLM
from litegen import completion
from litegen import ModelType


class Steps(BaseModel):
    steps: List[str]


class ReasonScore(BaseModel):
    reason: str
    score: int


class EvalLiteModel(DeepEvalBaseLLM):
    def __init__(
        self,
        model: Optional[ModelType] = None,
        *args,
        **kwargs,
    ):
        self.model_kwargs = kwargs
        self.model_name = model if model else "default"
        self.args = args
        self.kwargs = kwargs
        self.evaluation_cost = 0  # Initialize evaluation cost
        self._llm_func = kwargs.get("llm_func", None)
        super().__init__(self.model_name)

    def load_model(self):
        return self.model_name

    def llm_func(self, prompt, model) -> str:
        if self._llm_func:
            return self._llm_func(prompt=prompt,
                                  model=model,
                                  **self.model_kwargs)
        result = completion(model=model,
                            prompt=prompt,
                            **self.model_kwargs).choices[0].message.content
        return result

    def _parse_ai_response(self, response: str) -> dict:
        """Convert AI response to JSON format with better error handling"""
        try:
            # Remove markdown code block if present
            response = response.replace('```json', '').replace('```', '').strip()
            return json.loads(response)
        except json.JSONDecodeError:
            try:
                import re
                json_match = re.search(r'\{.*\}', response, re.DOTALL)
                if json_match:
                    return json.loads(json_match.group())
                return self._create_structured_response(response)
            except:
                return self._create_structured_response(response)

    def _create_structured_response(self, response: str) -> dict:
        """Create structured response based on content type"""
        lines = [line.strip() for line in response.split('\n') if line.strip()]

        # Try to identify steps format
        if any(line.startswith(str(i) + '.') for i in range(1, 10) for line in lines):
            steps = [line for line in lines if any(line.startswith(str(i) + '.') for i in range(1, 10))]
            return {"steps": steps}

        # Try to identify score/reason format
        import re
        score_match = re.search(r'\b([0-9]|10)\b', response)
        if score_match:
            score = int(score_match.group())
            reason = response.replace(score_match.group(), '').strip()
            return {"score": score, "reason": reason}

        return {"content": response}

    def _process_response(self, response: str, schema: Optional[Type[BaseModel]] = None) -> Union[BaseModel, str]:
        """Process response according to schema with improved handling"""
        if not schema:
            return response


        parsed_response = self._parse_ai_response(response)

        if schema.__name__ == 'Reason':
            return Reason(reason=parsed_response["reason"])
        elif schema.__name__ == 'Verdicts':
            verdicsts = [AnswerRelvancyVerdict(
                verdict=v,
                reason=r
            ) for v,r in parsed_response['verdicts']]
            return Verdicts(verdicts=verdicsts)
        elif schema.__name__ == 'Statements':
            return Statements(statements=parsed_response['statements'])

        elif schema.__name__ == 'Steps':
            if isinstance(parsed_response, dict) and 'steps' in parsed_response:
                return Steps(steps=parsed_response['steps'])
            return Steps(steps=[str(response)])

        elif schema.__name__ == 'ReasonScore':
            if isinstance(parsed_response, dict) and 'score' in parsed_response and 'reason' in parsed_response:
                return ReasonScore(score=parsed_response['score'], reason=parsed_response['reason'])
            # Handle non-standard response format
            score = parsed_response.get('score', 5)
            reason = parsed_response.get('reason', str(response))
            return ReasonScore(score=score, reason=reason)

        return response

    def generate(self, prompt: str, schema: Optional[Type[BaseModel]] = None, **kwargs) -> Union[BaseModel, str]:
        """Synchronous generation with schema support"""
        response = self.llm_func(prompt=prompt, model=self.model_name)
        processed_response = self._process_response(response, schema)
        return processed_response

    async def a_generate(self, prompt: str, schema: Optional[Type[BaseModel]] = None, **kwargs) -> Union[
        BaseModel, str]:
        """Asynchronous generation"""
        response = self.llm_func(prompt=prompt, model=self.model_name)
        processed_response = self._process_response(response, schema)
        return processed_response

    def get_model_name(self):
        return self.model_name

if __name__ == '__main__':

    # Initialize metric with a specific model
    answer_relevancy_metric = AnswerRelevancyMetric(
        threshold=0.7,
        model=EvalLiteModel(model='qwen2.5-coder:7b-instruct',
                            gpu=True)
    )

    # Create a test case
    test_case = LLMTestCase(
        input="What if these shoes don't fit?",
        actual_output="We offer a 30-day full refund at no extra costs.",
        retrieval_context=["All customers are eligible for a 30 day full refund at no extra costs."]
    )

    # Run evaluation
    evaluate([test_case], [answer_relevancy_metric])



from pathlib import Path

def _load_md_files():
    current_dir = Path(__file__).parent
    md_files = current_dir.glob("*.md")

    # Dictionary to store module contents
    module_contents = {}

    for md_file in md_files:
        var_name = md_file.stem
        with open(md_file, 'r') as f:
            module_contents[var_name] = f.read().strip()

    return module_contents

# Load all MD files and add them to global namespace
globals().update(_load_md_files())

# Export all MD file variables
__all__ = [f.stem for f in Path(__file__).parent.glob("*.md")]



import asyncio
import re
from concurrent.futures import ThreadPoolExecutor
from dataclasses import field
from multiprocessing import cpu_count
from urllib.parse import quote_plus

from bs4 import BeautifulSoup
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from pydantic import BaseModel
from typing import List
import time


class SearchResult(BaseModel):
    query: str
    urls: list = field(default_factory=list)
    search_provider: str = ""
    page_source: str = ""


class OptimizedMultiQuerySearcher:
    def __init__(self, chromedriver_path="/usr/local/bin/chromedriver", max_workers=None,
                 animation=False):
        self.chromedriver_path = chromedriver_path
        self.max_workers = max_workers or min(32, cpu_count() - 2)
        self.animation = animation
        self.driver_pool = []
        self.params = {
            "bing": {"search": "b_results", "find_element": ".b_algo", "head_selector": "h2"},
            "google": {"search": "search", "find_element": "div.g", "head_selector": "h3"}
        }

    def _setup_driver(self):
        chrome_options = Options()
        if not self.animation:
            chrome_options.add_argument("--headless")
        chrome_options.add_argument("--disable-gpu")
        chrome_options.add_argument("--no-sandbox")
        chrome_options.add_argument("--disable-dev-shm-usage")
        chrome_options.add_argument("--disable-extensions")
        chrome_options.add_argument("--disable-images")
        chrome_options.add_argument("--blink-settings=imagesEnabled=false")
        chrome_options.add_argument("--start-maximized")
        chrome_options.page_load_strategy = 'none'
        chrome_options.add_argument("--disable-javascript")  # If you don't need JS

        service = Service(self.chromedriver_path)
        return webdriver.Chrome(service=service, options=chrome_options)

    def _get_driver(self):
        if not self.driver_pool:
            return self._setup_driver()
        return self.driver_pool.pop()

    def _return_driver(self, driver):
        self.driver_pool.append(driver)

    def cleanup(self):
        if hasattr(self, 'driver'):
            self.driver.quit()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()

    def search_single_query(self, query, num_results=10,
                            search_provider="google",
                            min_waiting_time: int = 2
                            ):
        driver = self._get_driver()
        try:
            encoded_query = quote_plus(query)
            search_url = f"https://www.{search_provider}.com/search?q={encoded_query}"
            driver.get(search_url)
            time.sleep(min_waiting_time)
            # Wait for initial search results container
            WebDriverWait(driver, 3).until(
                EC.presence_of_element_located((By.ID, self.params[search_provider]["search"]))
            )

            # Wait for search result links to be present and visible
            if search_provider == "google":
                WebDriverWait(driver, 3).until(
                    EC.presence_of_all_elements_located((By.CSS_SELECTOR, "div.g a[href^='https://']"))
                )
            else:  # bing
                WebDriverWait(driver, 3).until(
                    EC.presence_of_all_elements_located((By.CSS_SELECTOR, ".b_algo a[href^='https://']"))
                )

            driver.execute_script("return document.readyState") == "complete"

            soup = BeautifulSoup(driver.page_source, 'html.parser')
            atags = soup.find_all('a',
                                  attrs={'href': re.compile("^https://"),
                                         # 'aria-label': True
                                         })

            urls = self.extract_urls(atags)

            return SearchResult(
                query=query,
                urls=urls,
                search_provider=search_provider,
                page_source=driver.page_source
            )
        except Exception as e:
            print(f'exception as {e}')
            return SearchResult(query=query)
        finally:
            self._return_driver(driver)

    async def search_multiple_queries(self, queries: List[str], num_results=10,
                                      search_provider=None, return_only_urls=False,
                                      min_waiting_time: int = 2
                                      ) -> List[SearchResult]:
        loop = asyncio.get_event_loop()
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            providers_list = self.params.keys() if search_provider is None else [search_provider]
            tasks = [
                loop.run_in_executor(
                    executor,
                    self.search_single_query,
                    query,
                    num_results,
                    provider,
                    min_waiting_time
                )
                for query in queries for provider in providers_list
            ]
            result = await asyncio.gather(*tasks)
            if return_only_urls:
                all_urls = [url.urls for url in result]
                filtered_urls = [y for x in zip(*all_urls) for y in x]
                return filtered_urls
            return result


    def extract_urls(self, atags):
        all_urls = [u.get("href") for u in atags if u.get("href")]

        filtered_urls = []
        for url in all_urls:
            if 'google' not in url:
                filtered_urls.append(url)

        yt_urls = []
        non_yt_urls = []
        for url in filtered_urls:
            if 'youtube' in url:
                yt_urls.append(url)
            else:
                non_yt_urls.append(url)

        final_urls = non_yt_urls + yt_urls

        return final_urls


class RealTimeGoogleSearchProvider:
    def __init__(
        self,
        search_provider="google",
        chromedriver_path="/usr/local/bin/chromedriver",
        max_workers=None,
        animation=False
    ):
        self.search_provider = search_provider
        self.chromedriver_path = chromedriver_path
        self.max_workers = max_workers
        self.animation = animation

    def search(self, query: str, max_urls=50) -> List[str]:
        with OptimizedMultiQuerySearcher(chromedriver_path=self.chromedriver_path,
                                         max_workers=self.max_workers,
                                         animation=self.animation) as searcher:
            res = searcher.search_single_query(query, search_provider=self.search_provider)
            # print("@@@@@")
            # Path('goolge_page_source.txt').write_text(res.page_source)
            return res.urls[:max_urls]

    async def _async_batch_search(self, batch_queries, max_urls=50) -> List[str]:
        with OptimizedMultiQuerySearcher(chromedriver_path=self.chromedriver_path,
                                         max_workers=self.max_workers,
                                         animation=self.animation) as searcher:
            all_urls = await searcher.search_multiple_queries(
                queries=batch_queries,
                search_provider=self.search_provider
            )
            all_urls = [url.urls for url in all_urls]
            filtered_urls = [y for x in zip(*all_urls) for y in x]
            filtered_urls = [self._extract_until_hash(x) if self._is_hash(x) else x for x in filtered_urls]
            filtered_urls = [_ for _ in filtered_urls if _]
            duplicate_removed_urls = self._remove_duplicate_urls(filtered_urls)
            return duplicate_removed_urls[:max_urls]

    def search_batch(self, batch_queries, max_urls=50) -> List[str]:
        return asyncio.run(self._async_batch_search(batch_queries, max_urls=max_urls))

    def __call__(self, query: str | list, *args, **kwargs):
        if isinstance(query, str):
            return self.search(query, *args, **kwargs)
        return self.search_batch(query, *args, **kwargs)

    def _is_hash(self, x):
        return '#' in x

    def _extract_until_hash(self, x):
        results = re.findall(r'(.*)#', x)
        if results:
            return results[0]
        return ""

    def _remove_duplicate_urls(self, filtered_urls):
        """Remove duplicates  by maintaining order"""
        seen = set()
        seen_add = seen.add
        return [x for x in filtered_urls if not (x in seen or seen_add(x))]





from typing import List, Tuple, Union

from wordllama import WordLlama

__wllm = None
__dim = None


def __get_word_llama_model(dim: int = 256):
    global __wllm, __dim
    if __wllm is None or __dim != dim:
        __dim = dim
        __wllm = WordLlama.load(dim=dim)
    return __wllm


def wlsplit(text: str) -> List[str]:
    return __get_word_llama_model().split(text)


def wlanswer(context: str, query: str, k: int = 1):
    __chunks = wlsplit(context)
    return "\n".join(wltopk(__chunks, query, k=k))




def wltopk(docs: List[str], query: str, k: int = 3) -> List[str]:
    return __get_word_llama_model().topk(query, docs, k=k)


def wlembed(texts: Union[str, List[str]]):
    return __get_word_llama_model().embed(texts)


def wlsimchunks(context: str, query: str, k: int = 3):
    __chunks = wlsplit(context)
    return wltopk(__chunks, query, k=k)


def wlanswerm(query: str, context: str, k: int = 3):
    __chunks = wlsplit(context)
    return "\n".join(wltopk(__chunks, query, k=k))


def wlanswerl(query: str, context: str, k: int = 5):
    __chunks = wlsplit(context)
    return "\n".join(wltopk(__chunks, query, k=k))


def wlscore(text1: str, text2: str) -> float:
    return __get_word_llama_model().similarity(text1, text2)


def wlrank(query: str, candidates: List[str], sort: bool = True, batch_size: int = 64) -> List[Tuple[str, float]]:
    return __get_word_llama_model().rank(query, candidates, sort=sort, batch_size=batch_size)


def wldeduplicate(docs: List[str], return_indices: bool = False, threshold: float = 0.5) -> List[str]:
    return __get_word_llama_model().deduplicate(docs, return_indices=return_indices, threshold=threshold)


def wlcluster(docs: List[str], k: int, max_iterations: int = 100, tolerance: float = 1e-4, n_init: int = 3) -> Tuple[
    List[int], float]:
    return __get_word_llama_model().cluster(docs, k=k, max_iterations=max_iterations, tolerance=tolerance,
                                            n_init=n_init)


def wlfilter(query: str, docs: List[str], threshold: float = 0.3) -> List[str]:
    return __get_word_llama_model().filter(query, docs, threshold=threshold)
